from __future__ import annotations
from typing import List, Optional
import numpy as np
from PPVD.validation import validate_evenly_divisible, validate_matrix, validate_numpy_type, validate_tensor


def generate_raster(event_frames: List[List[int]], total_frames: Optional[int] = None) -> np.ndarray:
    """
    Generate raster from lists of frames containing an event (e.g., spikes)

    :param event_frames: list of event frames (e.g., spike frames)
    :type event_frames: list[list[int]]
    :param total_frames: total number of frames
    :type total_frames: Optional[int] = None
    :return: event matrix of neurons x total frames
    :rtype: numpy.ndarray
    """

    if not total_frames:  # if total frames not provided we estimate by finding the very last event
        total_frames = np.max([event for events in event_frames for event in events])+1  # + 1 to account for 0-index
    _neurons = len(event_frames)
    event_matrix = np.full((_neurons, total_frames), 0, dtype=np.int32)

    # Merge Here - could be done more efficiently but not priority
    for _neuron in range(_neurons):
        for _event in event_frames[_neuron]:
            event_matrix[_neuron, _event] = 1
    return event_matrix


@validate_matrix(pos=0)
@validate_evenly_divisible(numerator=0, denominator=1, axis=1)
def generate_tensor(traces_as_matrix: np.ndarray, chunk_size: int) -> np.ndarray:
    """
    Generates a tensor given chunk / trial indices

    :param traces_as_matrix: traces in matrix form (neurons x frames)
    :type traces_as_matrix: numpy.ndarray
    :param chunk_size: size of each chunk
    :type chunk_size: int
    :return: traces_as_tensor
    :rtype: numpy.ndarray
    """
    return np.stack(np.hsplit(traces_as_matrix, traces_as_matrix.shape[1]//chunk_size), axis=0)


@validate_tensor(pos=0)
def merge_tensor(traces_as_tensor: np.ndarray) -> np.ndarray:
    """
    Concatenate multiple trials or tiffs into single matrix:


    :param traces_as_tensor: chunk (trial, tiff, etc) x neurons x frames
    :type traces_as_tensor: numpy.ndarray
    :return: traces in matrix form
    :rtype: numpy.ndarray
    """
    return np.hstack(traces_as_tensor)


@validate_numpy_type(required_dtype="object", pos=0)
def merge_factorized_matrices(factorized_traces: np.ndarray, component: int = 0) -> np.ndarray:
    """
    Concatenate a neuron x chunk or trial array in which each element is a component x frame factorization of the
    original trace:


    :param factorized_traces: neurons x chunks (trial, tiff, etc) containing the neuron's trace factorized
        into several components
    :type factorized_traces: numpy.ndarray
    :param component: specific component to extract
    :type component: int
    :return: traces of specific component in matrix form
    :rtype: numpy.ndarray
    """
    _neurons, _trials = factorized_traces.shape
    _frames = np.concatenate(factorized_traces[0], axis=1)[0, :].shape[0]
    traces_as_matrix = np.full((_neurons, _frames), 0, dtype=factorized_traces.dtype)  # pre-allocated

    # Merge Here - could be done more efficiently but not priority
    for _neuron in range(_neurons):
        traces_as_matrix[_neuron, :] = np.concatenate(factorized_traces[_neuron], axis=1)[component, :]

    return traces_as_matrix



import pytest
import numpy as np
from CalSciPy.event_processing import bin_events, calculate_firing_rates, calculate_mean_firing_rates, \
    gaussian_smooth_firing_rates, normalize_firing_rates
from PPVD.style import TerminalStyle
from itertools import product

# samples
samples_event_matrix = np.full((5, 30), 1)
samples_event_matrix[0, :] = np.arange(30)
samples_event_matrix[1, np.arange(0, 30, 3)] = 3
samples_event_matrix[2, :] = np.arange(30)
samples_event_matrix[1, np.arange(0, 30, 2)] = 2
samples_event_matrix[4, :] = np.arange(30)
bin_length_ = [2, 3, 5, 10]
sample_binning = tuple(list(product([samples_event_matrix], bin_length_)))


@pytest.mark.parametrize(("matrix", "bin_lengths"), [*sample_binning])
def test_bin_events_passes(matrix, bin_lengths):
    binned_matrix = bin_events(matrix, bin_lengths)
    if binned_matrix.shape[0] != matrix.shape[0]:
        raise AssertionError(f"{TerminalStyle.YELLOW} Binned matrix lost features {TerminalStyle.RESET}")
    if not binned_matrix.shape[1] < matrix.shape[1]:
        raise AssertionError(f"{TerminalStyle.YELLOW} Binned matrix is the same shape as the original matrix "
                             f"{TerminalStyle.RESET}")


def test_calculate_firing_rates():
    pass


def test_mean_firing_rates():
    pass


def test_gaussian_smooth_firing_rates():
    pass


def test_normalize_firing_rates():
    pass



import subprocess
import sys
import os
from PPVD.style import TerminalStyle
import pytest
import toml
import importlib


def retrieve_details(path):
    details = toml.load(path).get("project")
    name = details.get("name")
    version = details.get("version")
    dependencies = details.get("dependencies")
    return name, version, dependencies


pyproject_file = "".join([os.getcwd(), "\\pyproject.toml"])


package_name, package_version, package_dependencies = retrieve_details(pyproject_file)

print(f"\n\n{TerminalStyle.ORANGE}Testing...{TerminalStyle.RESET}\n")
print(f"{TerminalStyle.YELLOW}Package: {TerminalStyle.BLUE}{package_name}{TerminalStyle.RESET}\n")
print(f"{TerminalStyle.YELLOW}Version: {TerminalStyle.BLUE}{package_version}{TerminalStyle.RESET}\n")
print(f"{TerminalStyle.YELLOW}Dependencies: {TerminalStyle.BLUE}{package_dependencies}{TerminalStyle.RESET}\n")


@pytest.mark.parametrize("path", [pyproject_file])
def test_install(path):
    subprocess.check_call([sys.executable, "-m", "pip", "install", "-e ."])



import pytest
import numpy as np
import os

from PPVD.style import TerminalStyle
from CalSciPy.reorganization import generate_raster, generate_tensor, merge_factorized_matrices, merge_tensor


# arguments for tests
sample_event_lists = (
    [([[1, 2, 3], [5, 6, 7], [0, 10, 20]], None),
     ([[1, 2, 3], [5, 6, 7], [0, 10, 20]], 21)]
)
sample_factorized_matrix = np.load("".join([os.getcwd(),
                                            "\\testing_data\\sample_variables\\sample_factorized_matrices.npy"]),
                                   allow_pickle=True)

sample_matrix = np.full((5, 100), 1)
sample_matrix[0, :] = np.arange(100)

sample_tensor = np.array([np.full((4, 5), 1), np.full((4, 5), 2), np.full((4, 5), 3)])


@pytest.mark.parametrize("matrix", [sample_matrix])
def test_generate_tensor_passes(matrix):
    tensor = generate_tensor(matrix, 25)
    if tensor.shape != (4, 5, 25):
        raise AssertionError(f"{TerminalStyle.YELLOW}: "
                             f"Generating tensor did not maintain correct shape: "
                             f"expected {(4, 5, 25)} received {TerminalStyle.BLUE}{tensor.shape}"
                             f"{TerminalStyle.RESET}")
    for _chunk in range(tensor.shape[0]):
        np.testing.assert_array_equal(tensor[_chunk, 0, :], np.arange(25)+(_chunk*25),
                                      f"{TerminalStyle.GREEN}Chunk {_chunk}: {TerminalStyle.YELLOW}"
                                      f"Generating tensor did not maintain correct order {TerminalStyle.RESET}")


def test_generate_tensor_fails():
    # validate not matrix
    with pytest.raises(AssertionError):
        generate_tensor(np.full((5, 90, 5), 1), 25)
    # validate not evenly divisible
    with pytest.raises(AssertionError):
        generate_tensor(np.full((5, 90), 1), 25)


@pytest.mark.parametrize(("event_frames", "total_frames"), [sample_event_lists[0], sample_event_lists[1]])
def test_generate_raster_passes(event_frames, total_frames):
    event_matrix = generate_raster(event_frames, total_frames)
    if event_matrix.shape[0] != 3:
        raise AssertionError(f"{TerminalStyle.YELLOW}Generate Raster did not generate the correct number of neurons: "
                             f"Input = {TerminalStyle.BLUE}{event_frames} {TerminalStyle.YELLOW}given"
                             f"{TerminalStyle.BLUE}{total_frames}{TerminalStyle.YELLOW} frames{TerminalStyle.RESET}")

    if event_matrix.shape[1] != 21:
        raise AssertionError(f"{TerminalStyle.YELLOW}Generate Raster did not generate the correct total frames: "
                             f"Input = {TerminalStyle.BLUE}{event_frames} {TerminalStyle.YELLOW}given"
                             f"{TerminalStyle.BLUE}{total_frames}{TerminalStyle.YELLOW} frames{TerminalStyle.RESET}")

    np.testing.assert_array_equal(np.sum(event_matrix, axis=0),
                                  np.array([1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]),
                                  err_msg=f"{TerminalStyle.YELLOW}Generate Raster did not correctly place the events"
                             f"Input = {TerminalStyle.BLUE}{event_frames} {TerminalStyle.YELLOW}given"
                             f"{TerminalStyle.BLUE}{total_frames}{TerminalStyle.YELLOW} frames and returned"
                                          f"{TerminalStyle.BLUE}{np.sum(event_matrix, axis=0)}{TerminalStyle.YELLOW}"
                                          f"instead of {TerminalStyle.BLUE}"
                                          f"{np.array([1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1])}"
                                          f"{TerminalStyle.RESET}")


@pytest.mark.parametrize("traces_as_tensor", [sample_tensor])
def test_merge_tensor_passes(traces_as_tensor):
    traces_as_matrix = merge_tensor(traces_as_tensor)
    if traces_as_matrix.shape != (4, 15):
        raise AssertionError(f"{TerminalStyle.YELLOW}Merge Tensor did not maintain correct shape:"
                             f"expected {(4, 15)} received {TerminalStyle.BLUE}{traces_as_matrix.shape}"
                             f"{TerminalStyle.RESET}")


def test_merge_tensor_fails():
    # validate tensor
    with pytest.raises(AssertionError):
        merge_tensor(np.full((5, 100), 1))


@pytest.mark.parametrize("factorized_matrix", [sample_factorized_matrix])
def test_merge_factorized_matrices_passes(factorized_matrix):

    for _component in range(3):
        component_matrix = merge_factorized_matrices(factorized_matrix, component=_component)
        if component_matrix.shape != (10, 450):
            raise AssertionError(f"{TerminalStyle.GREEN}Component {_component}{TerminalStyle.YELLOW}: "
                                 f"Merging factorized matrices did not maintain correct shape: "
                                 f"expected {(10, 450)} received {TerminalStyle.BLUE}{component_matrix.shape}"
                                 f"{TerminalStyle.RESET}")
        np.testing.assert_array_equal(component_matrix[_component, :], np.concatenate([np.arange(100)*_component,
                                                                                 np.arange(100)*_component,
                                                                                 np.arange(100)*_component,
                                                                                 np.arange(100) * _component,
                                                                                 np.arange(50)*_component]),
                                      f"{TerminalStyle.GREEN}Component {_component}{TerminalStyle.YELLOW}: "
                                      f"Merging factorized matrices did not maintain correct order"
                                      f"{TerminalStyle.RESET}")


def test_merge_factorized_matrices_fails():
    with pytest.raises(TypeError):
        merge_factorized_matrices(np.full((5, 90), 1, dtype=np.float64))



from __future__ import annotations
import numpy as np
import os
from tqdm.auto import tqdm
from tqdm import tqdm as tq
import tifffile
from typing import Tuple, Optional, Union
import math
import pathlib
from imageio import mimwrite

from PPVD.validation import validate_exists, validate_extension, validate_filename, validate_path
from PPVD.parsing import convert_permitted_types_to_required, if_dir_append_filename, if_dir_join_filename, \
    require_full_path, find_num_unique_files_given_static_substring, find_num_unique_files_containing_tag


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=0)
@validate_path(pos=0)
@validate_exists(pos=0)
def load_all_tiffs(folder: Union[str, pathlib.Path]) -> np.ndarray:
    """
    Loads all .tif's within a folder into a single numpy array. Assumes .tif files are the standard unsigned 16-bit
    exported by the majority (all?) of imaging software.

    :param folder: folder containing a sequence of tiff stacks
    :type folder: str or pathlib.Path
    :return: a numpy array containing the images (frames, y-pixels, x-pixels)
    :rtype: numpy.ndarray
    """
    if isinstance(folder, pathlib.Path):
        folder = str(folder)

    _filenames = [str(_filename.name) for _filename in pathlib.Path(folder).glob("*") if ".tif" in _filename.suffix]
    y_pix, x_pix = tifffile.TiffFile(folder + "\\" + _filenames[0]).pages[0].shape
    _num_frames = []  # initialize
    [_num_frames.append(len(tifffile.TiffFile(folder + "\\" + _filename).pages)) for _filename in _filenames]
    _total_frames = sum(_num_frames)
    complete_image = np.full((_total_frames, y_pix, x_pix), 0, dtype=np.uint16)
    _last_frame = 0

    for _filename in tqdm(
            range(len(_filenames)),
            total=len(_filenames),
            desc="Loading Images...",
            disable=False,
    ):
        complete_image[_last_frame:_last_frame+_num_frames[_filename], :, :] = \
            load_single_tiff(folder + "\\" + _filenames[_filename])
        _last_frame += _num_frames[_filename]

    return complete_image


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=0)
@validate_path(pos=0)
@if_dir_join_filename(default_name="video_meta.txt", flag_pos=0)
@validate_extension(required_extension=".txt", pos=0)
@validate_exists(pos=0)
def load_binary_meta(path: Union[str, pathlib.Path]) -> Tuple[int, int, int, str]:
    """
    Loads the meta file for an associated binary video

    :param path: The meta file (.txt ext) or a directory containing metafile
    :type path: str or pathlib.Path
    :return: A tuple where (frames, y-pixels, x-pixels, :class:`numpy.dtype`)
    :rtype: tuple[int, int, int, str]
    """
    _num_frames, _y_pixels, _x_pixels, _type = np.genfromtxt(path, delimiter=",", dtype="str")
    return int(_num_frames), int(_y_pixels), int(_x_pixels), str(_type)


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=0)
@validate_filename(pos=0)
@if_dir_append_filename(default_name="video_meta.txt", flag_pos=0)
@if_dir_join_filename(default_name="binary_video", flag_pos=0)
@validate_path(pos=1)
@validate_exists(pos=0)
@validate_exists(pos=1)
def load_mapped_binary(path: Union[str, pathlib.Path], meta_filename: Optional[str] = None, **kwargs: str) -> np.memmap:
    """
    Loads a raw binary file as numpy array without loading into memory (memmap). Enter a directory to autogenerate the
    default filenames (binary_video, video_meta.txt)

    :param path: absolute filepath for binary video or a folder containing a binary video with the default filename
    :type path: str or pathlib.Path
    :param meta_filename: absolute path to meta file
    :type meta_filename: Optional[str] = None
    :keyword mode: mode used in loading numpy.memmap (str, default = "r")
    :return: memmap (numpy) array (frames, y-pixels, x-pixels)
    :rtype: numpy.memmap
    """

    _mode = kwargs.get("mode", "r")

    _num_frames, _y_pixels, _x_pixels, _type = load_binary_meta(meta_filename)

    return np.memmap(path, dtype=_type, shape=(_num_frames, _y_pixels, _x_pixels), mode=_mode)
# TODO UNIT TEST FOR EXCEPTIONS


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=0)
@validate_filename(pos=0)
@if_dir_append_filename(default_name="video_meta.txt", flag_pos=0)
@if_dir_join_filename(default_name="binary_video", flag_pos=0)
@validate_path(pos=1)
@validate_exists(pos=0)
@validate_exists(pos=1)
def load_raw_binary(path: Union[str, pathlib.Path], meta_filename: Optional[str] = None) -> np.ndarray:
    """
    Loads a raw binary file as a numpy array. Enter a directory to autogenerate the default
    filenames (binary_video, video_meta.txt)

    :param path: absolute filepath for binary video or directory containing a file named binary video
    :type path: str or pathlib.Path
    :param meta_filename: absolute path to meta file
    :type meta_filename: Optional[str] = None
    :return: numpy array (frames, y-pixels, x-pixels)
    :rtype: numpy.ndarray
    """

    _num_frames, _y_pixels, _x_pixels, _type = np.genfromtxt(meta_filename, delimiter=",", dtype="str")
    _num_frames = int(_num_frames)
    _x_pixels = int(_x_pixels)
    _y_pixels = int(_y_pixels)
    return np.reshape(np.fromfile(path, dtype=_type), (_num_frames, _y_pixels, _x_pixels))
# TODO UNIT TEST FOR EXCEPTIONS


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=0)
@validate_path(pos=0)
@validate_exists(pos=0)
def load_single_tiff(path: Union[str, pathlib.Path]) -> np.ndarray:
    """
    Load a single .tif as a numpy array

    :param path: absolute filename
    :type path: str or pathlib.Path
    :return: numpy array (frames, y-pixels, x-pixels)
    :rtype: numpy.ndarray
    """
    image = []
    with tifffile.TiffFile(path) as _file:
        for _frame in _file.pages:
            image.append(_frame.asarray())

    return np.array(image)


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=1)
@validate_path(pos=1)
@require_full_path(pos=1)
@validate_extension(required_extension=".tif", pos=1)
def save_single_tiff(images: np.ndarray, path: Union[str, pathlib.Path], type_: Optional[np.dtype] = np.uint16) -> None:
    """
    Save a numpy array to a single .tif file. Size must be <4 GB.

    :param images: numpy array [frames, y pixels, x pixels]
    :type images: numpy.ndarray
    :param path: filename or absolute path
    :type path: str or pathlib.Path
    :param type_: type for saving
    :type type_: Optional[numpy.dtype] = numpy.uint16
    :rtype: None
    """

    # just save if single frame
    if len(images.shape) == 2:
        with tifffile.TiffWriter(path) as tif:
            tif.save(np.floor(images).astype(type_))
        return

    with tifffile.TiffWriter(path) as tif:
        for frame in np.floor(images).astype(type_):
            tif.save(frame)


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=1)
@validate_path(pos=1)
def save_tiff_stack(images: str, output_folder: Union[str, pathlib.Path],
                    type_: Optional[np.dtype] = np.uint16) -> None:
    """
    Save a numpy array to a sequence of .tif stacks

    :param images: numpy array (frames, y-pixels, x-pixels)
    :type images: numpy.ndarray
    :param output_folder: folder to save the sequence of .tif stacks
    :type output_folder: str or pathlib.Path
    :param type_: type for saving
    :type type_: Optional[numpy.dtype] = numpy.uint16
    :rtype: None
    """
    _num_frames = images.shape[0]

    _chunks = math.ceil(_num_frames / 7000)

    c_idx = 1
    for _chunk in range(0, _num_frames, 7000):

        _start_idx = _chunk
        _end_idx = _chunk + 7000
        if _end_idx > _num_frames:
            _end_idx = _num_frames + 1

        if c_idx < 10:
            save_single_tiff(images[_start_idx:_end_idx, :, :],
                             output_folder + "\\" + "Video_0" + str(c_idx) + "_of_" + str(
                                       _chunks) + ".tif", type_)
        else:
            save_single_tiff(images[_start_idx:_end_idx, :, :],
                             output_folder + "\\" + "Video_" + str(c_idx) + "_of_" + str(
                                       _chunks) + ".tif", type_)
        c_idx += 1

    return print("Finished Saving Tiffs")
# REFACTOR at some point the CHUNKING


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=1)
@validate_path(pos=1)
@if_dir_append_filename(default_name="video_meta.txt", flag_pos=1)
@if_dir_join_filename(default_name="binary_video", flag_pos=1)
@validate_extension(required_extension=".txt", pos=2)
def save_raw_binary(images: np.ndarray, path: Union[str, pathlib.Path],
                    meta_filename: Optional[Union[str, pathlib.Path]]) -> None:
    """
    Save a numpy array as a binary file with an associated meta .txt file

    :param images: numpy array (frames, y-pixels, x-pixels)
    :type images: numpy.ndarray
    :param path:  folder to save in or an absolute filepath for binary video file
    :type path: str
    :param meta_filename: absolute filepath for saving meta .txt file
    :type meta_filename: str
    :rtype: None
    """

    try:
        assert(pathlib.Path(path).parent.exists())
    except AssertionError:
        os.makedirs(str(pathlib.Path(path).parent))
    finally:
        with open(meta_filename, 'w') as f:
            f.writelines([str(images.shape[0]), ",", str(images.shape[1]), ",",
                          str(images.shape[2]), ",", str(images.dtype)])
    images.tofile(path)
    print("Finished saving images as a binary file.")
# TODO UNIT TEST FOR EXCEPTIONS


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=1)
@validate_path(pos=1)
@if_dir_join_filename(default_name="video.mp4", flag_pos=1)
@validate_extension(required_extension=".mp4", pos=1)
def save_video(images: np.ndarray, path: Union[str, pathlib.Path], fps: float = 30.0) -> None:
    """
    Save numpy array as an .mp4. Will be converted to uint8 if any other datatype.

    :param images: numpy array (frames, y-pixels, x-pixels)
    :type images: numpy.array
    :param path: absolute filepath or filename
    :type path: str or pathlib.Path
    :param fps: frame rate for saved video
    :type fps: float = 30.0
    :rtype: None
    """

    if images.dtype.type != np.uint8:
        print("\nForcing to unsigned 8-bit\n")
        images = images.astype(np.uint8)

    print("\nWriting Images to .mp4...\n")
    mimwrite(path, images, fps=fps, quality=10, macro_block_size=4)
    print("\nFinished writing images to .mp4.\n")
# TODO: NO UNIT TEST



from __future__ import annotations
import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter1d
from itertools import product


def bin_events(matrix: np.ndarray, bin_length: int) -> np.ndarray:
    """
    Bin events (e.g., spikes) using specified bin length

    :param matrix: matrix of n features x m samples
    :type matrix: numpy.ndarray
    :param bin_length: length of bin
    :type bin_length: int
    :return: binned_matrix of n features x m bins
    :rtype: numpy.ndarray
    """
    _features, _frames = matrix.shape
    _bins = pd.interval_range(0, _frames, freq=bin_length)
    binned_matrix = np.empty((_features, _frames // bin_length), dtype = np.float64)
    for _feature, _bin in product(range(_features), range(len(_bins))):
        binned_matrix[_feature, _bin] = \
            np.sum(matrix[_feature, int(_bins.values[_bin].left):int(_bins.values[_bin].right)])
    return binned_matrix


def calculate_firing_rates(spike_probability_matrix: np.ndarray, frame_rate: float = 30, in_place: bool = False) \
        -> np.ndarray:
    """
    Calculate firing rates

    :param spike_probability_matrix: matrix of n neuron x m samples where each element is the probability of a spike
    :type spike_probability_matrix: numpy.ndarray
    :param frame_rate: frame rate of dataset
    :type frame_rate: float = 30
    :param in_place: boolean indicating whether to perform calculation in-place
    :type in_place: bool = False
    :return: firing matrix of n neurons x m samples where each element is a binary indicating presence of spike event
    :rtype: numpy.ndarray
    """
    if in_place:
        firing_matrix = spike_probability_matrix
    else:
        firing_matrix = spike_probability_matrix.copy()

    firing_matrix *= frame_rate

    return firing_matrix
# TODO UNIT TEST


def calculate_mean_firing_rates(firing_matrix: np.ndarray) -> np.ndarray:
    """
    Calculate mean firing rate

    :param firing_matrix: matrix of n neuron x m samples where each element is either a spike or an
        instantaneous firing rate
    :type firing_matrix: numpy.ndarray
    :return: 1-D vector of mean firing rates
    :rtype: numpy.ndarray
    """
    return np.nanmean(firing_matrix, axis=1)
# TODO UNIT TEST


def gaussian_smooth_firing_rates(firing_matrix: np.ndarray, sigma: float, in_place: bool = False) -> np.ndarray:
    """
    Normalize firing rates using a 1-D gaussian filter

    :param firing_matrix: matrix of n neuron x m samples where each element is either a spike or an
        instantaneous firing rate
    :type firing_matrix: numpy.ndarray
    :param sigma: standard deviation of gaussian kernel
    :type sigma: float
    :param in_place: boolean indicating whether to perform calculation in-place
    :type in_place: bool = False
    :return: gaussian-smoothed firing rate matrix of n neurons x m samples
    :rtype: numpy.ndarray
    """
    if not in_place:
        return gaussian_filter1d(firing_matrix, sigma, axis=1)

    return gaussian_filter1d(firing_matrix, sigma, axis=1, output=firing_matrix)
# TODO UNIT TEST


def normalize_firing_rates(firing_matrix: np.ndarray, in_place: bool = False) -> np.ndarray:
    """
    Normalize firing rates by scaling to a max of 1.0. Non-negativity constrained.

    :param firing_matrix: matrix of n neuron x m samples where each element is either a spike or an
        instantaneous firing rate
    :type firing_matrix: numpy.ndarray
    :param in_place: boolean indicating whether to perform calculation in-place
    :type in_place: bool = False
    :return: normalized firing rate matrix of n neurons x m samples
    :rtype: numpy.ndarray
    """
    if in_place:
        normalized_matrix = firing_matrix
    else:
        normalized_matrix = firing_matrix.copy()

    normalized_matrix /= np.max(normalized_matrix, axis=0)
    normalized_matrix[normalized_matrix <= 0] = 0
    return normalized_matrix
# TODO UNIT TEST



# noinspection DuplicatedCode
from __future__ import annotations
import itertools
import numpy as np
from PIL import Image
from tqdm.auto import tqdm
from tqdm import tqdm as tq
from typing import Tuple, Optional, Union
import math
import pathlib
from prettytable import PrettyTable
from collections.abc import Iterable

from .io_tools import save_single_tiff
from PPVD.validation import validate_exists, validate_path
from PPVD.parsing import convert_permitted_types_to_required, find_num_unique_files_given_static_substring, \
    find_num_unique_files_containing_tag


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=pathlib.Path)
@validate_path(pos=0)
@validate_exists(pos=0)
def determine_bruker_folder_contents(folder: Union[str, pathlib.Path]) -> Tuple[int, int, int, int, int]:
    """
    This function determines the number of channels and planes within a folder containing .tif files
    exported by Bruker's Prairieview software. It also determines the size of the images (frames, y-pixels, x-pixels)

    :param folder: folder containing bruker imaging data
    :type folder: str or pathlib.Path
    :returns: channels, planes, frames, height, width
    :rtype: tuple[int, int, int, int, int]
    """

    _files = [_file for _file in folder.glob("*.tif") if _file.is_file()]

    def find_channels() -> int:
        nonlocal _files
        return find_num_unique_files_containing_tag("Ch", _files)

    def is_multiplane() -> bool:
        nonlocal _files
        if find_num_unique_files_containing_tag("Cycle0000", _files) > 1:
            return True
        else:
            return False

    def find_planes() -> int:
        nonlocal channels
        nonlocal _files
        if is_multiplane():
            return find_num_unique_files_given_static_substring("Cycle00001", _files)//channels
        else:
            return 1

    def find_frames() -> int:
        nonlocal channels
        if is_multiplane():
            return find_num_unique_files_given_static_substring("000001.ome", _files)//channels
        else:
            return find_num_unique_files_given_static_substring("Cycle00001", _files)//channels

    def find_dimensions() -> Tuple[int, int]:
        nonlocal folder
        nonlocal _files
        return np.asarray(Image.open("".join(str(_files[0])))).shape

    channels = find_channels()

    # noinspection PyTypeChecker
    return (channels, find_planes(), find_frames(), *find_dimensions())
# REFACTOR at some point the function string parsing calls for simplicity\extensibility


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=pathlib.Path, pos=0)
@validate_path(pos=0)
@validate_exists(pos=0)
def load_bruker_tiffs(folder: Union[str, pathlib.Path],
                      channels: Optional[int] = None, planes: Optional[int] = None) -> Tuple[np.ndarray]:
    """
    Load a sequence of .tif files from a directory containing .tif files exported by Bruker's Prairieview software to a
    numpy array. If multiple channels or multiple planes exist, each channel and plane combination is loaded to a
    separate numpy array.

    :param folder: folder containing a sequence of single frame tiff files
    :type folder: str or pathlib.Path
    :param channels: specific channel to load from dataset (zero-indexed)
    :type channels: Optional[int] = None
    :param planes: specific plane to load from dataset (zero-indexed)
    :type planes: Optional[int] = None
    :return: All .tif files in the directory loaded to a tuple of numpy arrays
        (frames, y-pixels, x-pixels, :class:`np.uint16`)
    :rtype: tuple[numpy.ndarray]
     """

    def load_images():
        nonlocal folder
        nonlocal tag
        nonlocal _frames
        nonlocal _y_pixels
        nonlocal _x_pixels
        nonlocal _tqdm_desc

        def load_single_image():
            nonlocal folder
            nonlocal _files
            nonlocal _file
            return np.asarray(Image.open(str(_files[_file])))

        tag = "".join(["*", tag, "*"])
        _files = [_file for _file in folder.glob("*.tif") if _file.match(tag)]
        if len(_files) == 0:
            return
        # Bruker is usually saved to .tif in 0-65536 (uint16) even though recording says 8192 (uint13)
        complete_image = np.full((_frames, _y_pixels, _x_pixels), 0, dtype=np.uint16)
        for _file in tqdm(
                range(_frames),
                total=_frames,
                desc=_tqdm_desc,
                disable=False,
        ):
            complete_image[_file, :, :] = load_single_image()

        return complete_image

    _channels, _planes, _frames, _y_pixels, _x_pixels = determine_bruker_folder_contents(folder)
    pretty_print_image_description(_channels, _planes, _frames, _y_pixels, _x_pixels)

    if channels is None:
        channels = range(2)
    if planes is None:
        planes = range(_planes)

    images = []  # append all channel/plane combos to this list

    # if planes is 1 bruker uses "ChX_XXXXXX" designator channel+frames
    if _planes == 1:
        if not isinstance(channels, Iterable):
            _tqdm_desc = "".join(["Loading Plane ", str(planes),
                                  " Channel ", str(channels), " images..."])
            tag = "".join(["Ch", str(channels + 1)])
            images.append(load_images())
            images = filter(lambda image: image is not None, images)
            return tuple(images)
        for _channel, _plane in itertools.product(channels, planes):
            _tqdm_desc = "".join(["Loading Plane ", str(_plane),
                                  " Channel ", str(_channel), " images..."])
            tag = "".join(["Ch", str(_channel+1)])
            images.append(load_images())
        images = filter(lambda image: image is not None, images)
        return tuple(images)

    if not isinstance(planes, Iterable):
        if not isinstance(channels, Iterable):
            _tqdm_desc = "".join(["Loading Plane ", str(planes),
                                  " Channel ", str(channels), " images..."])
            tag = "".join(["Ch", str(channels + 1), "_00000", str(planes + 1)])
            images.append(load_images())
            images = filter(lambda image: image is not None, images)
            return tuple(images)
        for _channel in channels:
            _tqdm_desc = "".join(["Loading Plane ", str(planes),
                                  " Channel ", str(_channel), " images..."])
            tag = "".join(["Ch", str(_channel + 1), "_00000", str(planes + 1)])
            images.append(load_images())
        images = filter(lambda image: image is not None, images)
        return tuple(images)

    # if planes > 1 bruker uses "CycleXXXXX_ChX_XXXXXX" designator for frames+channel+plane
    for _channel, _plane in itertools.product(channels, planes):
        _tqdm_desc = "".join(["Loading Plane ", str(_plane),
                              " Channel ", str(_channel), " images..."])
        tag = "".join(["Ch", str(_channel+1), "_00000", str(_plane+1)])
        images.append(load_images())
    images = filter(lambda image: image is not None, images)
    return tuple(images)
# REFACTOR at some point


def pretty_print_image_description(channels, planes, frames, height, width) -> None:
    """
    Function prints the description of an imaging dataset as a table.

    :param channels: number of channels
    :type channels: int
    :param planes: number of planes
    :type planes: int
    :param frames: number of frames
    :type frames: int
    :param height: y-pixels
    :type height: int
    :param width: x-pixels
    :type width: int
    :rtype: None
    """
    _table = PrettyTable()
    _table.header = False
    _table.add_row(["Total Images Detected", channels * planes * frames])
    _table.add_row(["Channels", channels])
    _table.add_row(["Planes", planes])
    _table.add_row(["Frames", frames])
    _table.add_row(["Height", height])
    _table.add_row(["Width", width])
    print("\n")
    print(_table)


@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=0)
@convert_permitted_types_to_required(permitted=(str, pathlib.Path), required=str, pos=1)
@validate_path(pos=0)
@validate_path(pos=1)
@validate_exists(pos=0)
def repackage_bruker_tiffs(input_folder: Union[str, pathlib.Path], output_folder: Union[str, pathlib.Path],
                           *args: Union[int, tuple[int]]) -> None:
    """
    Repackages a folder containing .tif files exported by Bruker's Prairieview software into a sequence of <4 GB .tif
    stacks.

    :param input_folder: folder containing a sequence of single frame .tif files exported by Bruker's Prairieview
    :type input_folder: str or pathlib.Path
    :param output_folder: empty folder where .tif stacks will be saved
    :type output_folder: str or pathlib.Path
    :param args: optional argument to indicate the repackaging of a specific channel and/or plane
    :type args: int
    :rtype: None
    """

    # code here is pretty rough, needs TLC. ~horror~

    def load_image():
        nonlocal input_folder
        nonlocal _files
        nonlocal _file
        nonlocal _offset
        return np.asarray(Image.open(str(_files[_file + _offset])))

    def find_files(tag: Union[str, list[str]]):
        nonlocal input_folder
        nonlocal _files

        def check_file_contents(tag_: str, file_: pathlib.WindowsPath) -> bool:
            if tag_ in str(file_.stem).split("_"):
                return True
            else:
                return False

        # reset, rather maintain code as is then make a new temporary variable since this
        # is basically instant
        _files = [_file for _file in pathlib.Path(input_folder).glob("*.tif")]

        # now find
        if isinstance(tag, list):
            tag = "".join([tag[0], "_", tag[1]])
            _files = [_file for _file in _files if tag in str(_file.stem)]
        else:
            _files = [_file for _file in _files if check_file_contents(tag, _file)]

    def not_over_4gb() -> bool:
        nonlocal _files
        nonlocal _y
        nonlocal _x

        gb = np.full((_files.__len__(), _y, _x), 1, dtype=np.uint16).nbytes
        if gb <= 3.9:  # 3.9 as a safety buffer
            return True
        else:
            return False

    _files = [_file for _file in pathlib.Path(input_folder).rglob("*.tif")]
    _channels, _planes, _frames, _y, _x = determine_bruker_folder_contents(input_folder)
    pretty_print_image_description(_channels, _planes, _frames, _y, _x)

    # finding the files for a specific channel/plane here
    if args:
        # unpack if necessary
        if isinstance(args[0], tuple):
            _c = args[0][0]
            _p = args[0][1]
            args = tuple([_c, _p])

        if _channels > 1 and _planes > 1 and len(args) >= 2:
            _base_tag = "00000"
            _tag = ["".join(["Ch", str(args[0]+1)]), "".join([_base_tag, str(args[1]+1), ".ome"])]
            find_files(_tag)
        elif _channels == 1 and _planes > 1 and len(args) == 1:
            _base_tag = "00000"
            _tag = "".join([_base_tag, str(args[0]+1), ".ome"])
            find_files(_tag)
            print(_files)
        elif _channels == 1 and _planes > 1 and len(args) >= 2:
            _base_tag = "00000"
            _tag = "".join([_base_tag, str(args[1]+1), ".ome"])
            find_files(_tag)
        elif _channels > 1 and _planes == 1:
            _tag = "".join(["Ch", str(args[0]+1)])
            find_files(_tag)
        else:
            pass
    else:
        try:
            assert(_channels == 1)
        except AssertionError:
            print("Folder contains multiple channels")
            return
        try:
            assert(_planes == 1)
        except AssertionError:
            print("Folder contains multiple planes")
            return

    # noinspection PyTypeChecker

    # Decide whether saving in single stack is possible
    if not_over_4gb():
        _images = np.full((_frames, _y, _x), 0, dtype=np.uint16)
        _offset = 0
        for _file in range(_frames):
            _images[_file, :, :] = load_image()
        save_single_tiff(_images, "".join([output_folder, "\\compiledVideo_01_of_1.tif"]))
        return
    else:
        # noinspection PyTypeChecker
        _chunks = math.ceil(_frames/7000)
        c_idx = 1
        _offset = int()

        _pbar = tq(total=_frames)
        _pbar.set_description("Repackaging Bruker Tiffs...")

        for _chunk in range(0, _frames, 7000):

            _start_idx = _chunk
            _offset = _start_idx
            _end_idx = _chunk + 7000
            _chunk_frames = _end_idx-_start_idx
            # If this is the last chunk which may not contain a full 7000 frames...
            if _end_idx > _frames:
                _end_idx = _frames
                _chunk_frames = _end_idx - _start_idx
                _end_idx += 1

            image_chunk = np.full((_chunk_frames, _y, _x), 0, dtype=np.uint16)

            for _file in range(_chunk_frames):
                image_chunk[_file, :, :] = load_image()
                _pbar.update(1)

            if c_idx < 10:
                save_single_tiff(image_chunk, output_folder + "\\" + "compiledVideo_0" + str(c_idx) + "_of_" +
                                 str(_chunks) + ".tif")
            else:
                save_single_tiff(image_chunk, output_folder + "\\" + "compiledVideo_" + str(c_idx) + "_of_" +
                                 str(_chunks) + ".tif")
            c_idx += 1
        _pbar.close()
    return
# REFACTOR at some point



import os
import pytest
from shutil import rmtree
from CalSciPy.io_tools import load_all_tiffs, load_single_tiff, save_raw_binary, load_raw_binary, \
    save_single_tiff, save_tiff_stack, save_video, load_binary_meta
from CalSciPy.bruker import determine_bruker_folder_contents, repackage_bruker_tiffs, load_bruker_tiffs
import numpy as np
import pathlib
# noinspection PyProtectedMember
from PPVD.style import TerminalStyle

# noinspection DuplicatedCode
FIXTURE_DIR = "".join([os.getcwd(), "\\testing_data"])

DATASET = pytest.mark.datafiles(
    "".join([FIXTURE_DIR, "\\sample_datasets"]),
    keep_top_dir=False,
    on_duplicate="ignore",
)


def read_descriptions(file):
    return np.genfromtxt(str(file), delimiter=",", dtype="int")


@DATASET
def test_single_tiff_load_and_save(datafiles, tmp_path):
    # INGEST
    for _dir in datafiles.listdir():
        _input_image = next(pathlib.Path(_dir).glob("single.tif"))
        _descriptions = next(pathlib.Path(_dir).glob("description.txt"))
        _output_folder = "".join([str(tmp_path), "\\", str(pathlib.Path(_dir).stem), "_output"])
        # MAKE OUTPUT FOLDER
        os.mkdir(_output_folder)
        _output_file = "".join([str(tmp_path), "\\", str(pathlib.Path(_dir).stem), "_output\\single.tif"])
        # GET COMPARISON DESCRIPTIONS
        _descriptions = read_descriptions(_descriptions)
        # TEST
        _image = load_single_tiff(_input_image)
        np.testing.assert_array_equal(_image.shape, [1, *_descriptions[3:5]], err_msg=f"{TerminalStyle.GREEN}"
                                                                                      f"Image Mismatch: "
                                                                                      f"{TerminalStyle.YELLOW}"
                                                                                      f"failed on dataset " 
                                                                                      f"{TerminalStyle.BLUE}"
                                                                                      f"{pathlib.Path(_dir).name}"
                                                                                      f"{TerminalStyle.YELLOW}"
                                                                                      f" during first loading"
                                                                                      f"{TerminalStyle.RESET}")
        save_single_tiff(_image, _output_file)
        _image2 = load_single_tiff(_output_file)
        np.testing.assert_array_equal(_image, _image2, err_msg=f"{TerminalStyle.GREEN}Image Mismatch: "
                                                               f"{TerminalStyle.YELLOW}failed on dataset " 
                                                               f"{TerminalStyle.BLUE}{pathlib.Path(_dir).name}" 
                                                               f"{TerminalStyle.YELLOW} during second loading"
                                                               f"{TerminalStyle.RESET}")

    rmtree(tmp_path)


def test_single_tiff_fails():
    # LOAD
    with pytest.raises(ValueError):
        load_single_tiff("")  # FAIL VALIDATE PATH NO ROOT/DRIVE
    with pytest.raises(ValueError):
        load_single_tiff("C:\\&^6*")  # FAIL VALIDATE PATH PERMITTED CHARS
    with pytest.raises(FileNotFoundError):
        load_single_tiff("C:\\673469346349673496734967349673205-3258-32856-3486")
        # FAIL VALIDATE PATH NOT FOUND
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        load_single_tiff(125.6)
    # SAVE
    with pytest.raises(ValueError):
        save_single_tiff([1, 2, 3, 4, 5], "")  # FAIL VALIDATE PATH NO ROOT/DRIVE
    with pytest.raises(ValueError):
        save_single_tiff([1, 2, 3, 4, 5], "C:\\&^6*")  # FAIL VALIDATE PATH PERMITTED CHARS
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        save_single_tiff([1, 2, 3, 4, 5], 125.6)  # FAIL TYPE
    with pytest.raises(ValueError):
        save_single_tiff([1, 2, 3, 4, 5], "C:\\file.mp4")  # FAIL EXTENSION


@DATASET
def test_binaries_load_and_save(datafiles, tmp_path):
    for _dir in datafiles.listdir():
        # INGEST
        _input_image = next(pathlib.Path(_dir).glob("Video_01_of_1.tif"))
        _descriptions = next(pathlib.Path(_dir).glob("description.txt"))
        _output_folder = "".join([str(tmp_path), "\\", str(pathlib.Path(_dir).stem), "_output"])
        # MAKE OUTPUT FOLDER
        os.mkdir(_output_folder)
        # GET COMPARISON
        _descriptions = read_descriptions(_descriptions)
        # TEST
        _images1 = load_single_tiff(_input_image)
        np.testing.assert_array_equal(_images1.shape, _descriptions[2:5], err_msg=f"{TerminalStyle.GREEN} "
                                                                                  f"Image Mismatch: "
                                                                                  f"{TerminalStyle.YELLOW}"
                                                                                  f"failed on dataset "
                                                                                  f"{TerminalStyle.BLUE}"
                                                                                  f"{pathlib.Path(_dir).name} "
                                                                                  f"{TerminalStyle.YELLOW}"
                                                                                  f"during first loading"
                                                                                  f"{TerminalStyle.RESET}")
        save_raw_binary(_images1, _output_folder)
        _images2 = load_raw_binary(_output_folder)
        np.testing.assert_array_equal(_images1, _images2, err_msg=f"{TerminalStyle.GREEN}Image Mismatch: "
                                                                  f"{TerminalStyle.YELLOW}failed on dataset "
                                                                  f"{TerminalStyle.BLUE}{pathlib.Path(_dir).name} "
                                                                  f"{TerminalStyle.YELLOW} during second loading "
                                                                  f"{TerminalStyle.RESET}")

    rmtree(tmp_path)


def test_binary_load_meta_fails():
    with pytest.raises(ValueError):
        load_binary_meta("")  # FAIL VALIDATE PATH NO ROOT/DRIVE
    with pytest.raises(ValueError):
        load_binary_meta("C:\\&^6*")  # FAIL VALIDATE PATH PERMITTED CHARS
    with pytest.raises(FileNotFoundError):
        load_binary_meta("C:\\673469346349673496734967349673205-3258-32856-3486")
        # FAIL VALIDATE PATH NOT FOUND
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        load_binary_meta(125.6)
        # FAIL WRONG TYPE
    with pytest.raises(FileNotFoundError):
        load_binary_meta("C:\\file")  # No extension but adds then fails
    with pytest.raises(ValueError):
        load_binary_meta("C:\\file.mp4")  # No extension but catches


@DATASET
def test_tiff_stack_load_and_save(datafiles, tmp_path):
    for _dir in datafiles.listdir():
        # INGEST
        _input_image = next(pathlib.Path(_dir).glob("Video_01_of_1.tif"))
        _descriptions = next(pathlib.Path(_dir).glob("description.txt"))
        _output_folder = "".join([str(tmp_path), "\\", str(pathlib.Path(_dir).stem), "_output"])
        os.mkdir(_output_folder)
        # GET COMPARISON DESCRIPTIONS
        _descriptions = read_descriptions(_descriptions)
        # TEST
        _image1 = load_single_tiff(_input_image)
        np.testing.assert_array_equal(_image1.shape, _descriptions[2:5], err_msg=f"{TerminalStyle.GREEN}"
                                                                                 f"Image Mismatch: "
                                                                                 f"{TerminalStyle.YELLOW}"
                                                                                 f"failed on dataset "
                                                                                 f"{TerminalStyle.BLUE}"
                                                                                 f"{pathlib.Path(_dir).name} "
                                                                                 f"{TerminalStyle.YELLOW}"
                                                                                 f"during first loading "
                                                                                 f"{TerminalStyle.RESET}")
        save_tiff_stack(_image1, _output_folder)
        _image2 = load_all_tiffs(_output_folder)
        np.testing.assert_array_equal(_image1, _image2, err_msg=f"{TerminalStyle.GREEN}"
                                                                f"Image Mismatch: "
                                                                f"{TerminalStyle.YELLOW}"
                                                                f"failed on dataset "
                                                                f"{TerminalStyle.BLUE}"
                                                                f"{pathlib.Path(_dir).name} "
                                                                f"{TerminalStyle.YELLOW}"
                                                                f"during second loading "
                                                                f"{TerminalStyle.RESET}")


def test_tiff_stacks_fails():
    # LOAD
    with pytest.raises(ValueError):
        load_all_tiffs("")  # FAIL VALIDATE PATH NO ROOT/DRIVE
    with pytest.raises(ValueError):
        load_all_tiffs("C:\\&^6*")  # FAIL VALIDATE PATH PERMITTED CHARS
    with pytest.raises(FileNotFoundError):
        load_all_tiffs("C:\\673469346349673496734967349673205-3258-32856-3486")
        # FAIL VALIDATE PATH NOT FOUND
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        load_all_tiffs(125.6)
        # FAIL WRONG TYPE
    # SAVE
    with pytest.raises(ValueError):
        save_tiff_stack([1, 2, 3, 4, 5], "")  # FAIL VALIDATE PATH NO ROOT/DRIVE
    with pytest.raises(ValueError):
        save_tiff_stack([1, 2, 3, 4, 5], "C:\\&^6*")  # FAIL VALIDATE PATH PERMITTED CHARS
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        save_tiff_stack([1, 2, 3, 4, 5], 125.6)  # FAIL TYPE


@DATASET
def test_video_load_and_save(datafiles, tmp_path):
    for _dir in datafiles.listdir():
        # INGEST
        _input_file = next(pathlib.Path(_dir).glob("Video_01_of_1.tif"))
        _output_folder = "".join([str(tmp_path), "\\", str(pathlib.Path(_dir).stem), "_output"])
        _descriptions = next(pathlib.Path(_dir).glob("description.txt"))
        _descriptions = read_descriptions(_descriptions)
        # MAKE OUTPUT FOLDER
        os.mkdir(_output_folder)
        _image = load_single_tiff(_input_file)
        save_video(_image, _output_folder)
    rmtree(tmp_path)


def test_clean_up_directory(tmp_path):
    _tmp_dir_to_wipe = str(pathlib.Path(tmp_path).parent)
    rmtree(_tmp_dir_to_wipe)



from . import bruker
from . import coloring
from . import event_processing
from . import image_processing
from . import interactive_visuals
from . import io_tools
from . import reorganization
from . import static_visuals
from . import trace_processing
from . import validators
from .version import version


name = "CalSciPy"



from __future__ import annotations
from PPVD.parsing import parameterize
from PPVD.style import TerminalStyle
from typing import Callable, Tuple
from functools import wraps


@parameterize
def validate_longest_numpy_dimension(function: Callable, axis: int = 0, pos: int = 0) -> Callable:
    """
    Decorator for validating longest dimension of a numpy array
    (e.g., to ensure a video is organized in frame x height x width format)

    :param function: function to be decorated
    :type function: Callable
    :param axis: axis to be assert is the longest
    :type axis: int
    :param pos: index of the argument to be validated
    :type pos: int
    """

    @wraps(function)
    def decorator(*args, **kwargs):
        arg_shape = args[pos]
        long_axis = arg_shape[axis]
        axes_list = list(arg_shape)
        axes_list.pop(axis)

        for _axis in axes_list:
            if _axis > long_axis:
                raise AssertionError(f"{TerminaLStyle.GREEN}Input {pos} Improper Format: "
                                     f"{TerminalStyle.YELLOW} axis {TerminalStyle.BLUE}{axis}"
                                     f"{TerminalStyle.YELLOW} ought to be larger than axis "
                                     f"{TerminalStyle.BLUE}{_axis}{TerminalStyle.RESET}")
        # noinspection PyArgumentList
        return function(*args, **kwargs)

    return decorator


@parameterize
def validate_numpy_dimension_odd(function: Callable, odd_dimensions: Tuple[int] = tuple([0]), pos: int = 0) -> Callable:
    """
    Decorator for validating numpy dimension is odd

    :param function: function to be decorated
    :type function: Callable
    :param odd_dimensions: which dimensions odd_dimensions
    :type odd_dimensions: tuple[int]
    :param pos: index of the argument to be validated
    :type pos: int
    """
    @wraps(function)
    def decorator(*args, **kwargs):
        for _dim in odd_dimensions:
            if args[pos].shape[_dim] % 2 != 0:
                raise TypeError(f"{TerminalStyle.GREEN}Input {pos} Improper Format"
                            f"{TerminalStyle.YELLOW}the dimension "
                            f"{TerminalStyle.BLUE}{_dim}{TerminalStyle.YELLOW}"
                            f" must be odd{TerminalStyle.RESET}")
        # noinspection PyArgumentList
        return function(*args, **kwargs)
    return decorator



import pytest


def test_blockwise_fast_filter_tiff():
    pass


def test_clean_image_stack():
    pass


def test_fast_filter_images():
    pass


def test_filter_images():
    pass


def test_grouped_z_project():
    pass



from __future__ import annotations
from typing import Callable, List, Tuple, Sequence, Optional, Union, Any
import numpy as np
from tqdm.auto import tqdm
import scipy.ndimage
import skimage.measure
import math

from .validators import validate_longest_numpy_dimension, validate_numpy_dimension_odd

try:
    import cupy
    import cupyx.scipy.ndimage
except ModuleNotFoundError or ImportError:
    pass


@validate_longest_numpy_dimension(axis=0, pos=0)
@validate_numpy_dimension_odd(odd_dimensions=(0, 1, 2), pos=1)
def blockwise_fast_filter_tiff(images: np.ndarray, mask: np.ndarray = np.ones((3, 3, 3)), 
                               block_size: int = 21000, block_buffer: int = 500) -> np.ndarray:
    """
    GPU-parallelized multidimensional median filter performed in overlapping blocks.

    Designed for use on arrays larger than the available memory capacity.

    Footprint is of the form np.ones((frames, y pixels, x pixels)) with the origin in the cente

    :param images: images stack to be filtered
    :type images: numpy.ndarray
    :param mask: mask of the median filter
    :type mask: numpy.ndarray = np.ones((3, 3, 3))
    :param block_size: the size of each block. Must fit within memory
    :type block_size: int = 21000
    :param block_buffer: the size of the overlapping region between block
    :type block_buffer: int = 500
    :return: images: numpy array (frames, y pixels, x pixels)
    :rtype: numpy.ndarray
    """
    _total_frames = images.shape[0]
    _blocks = range(0, _total_frames, block_size)
    _num_blocks = len(_blocks)
    _remainder = np.full((block_buffer, images.shape[1], images.shape[2]), 0, dtype=np.int16)

    for _block in tqdm(
            range(_num_blocks),
            total=_num_blocks,
            desc="Filtering images...",
            disable=False,
    ):
        if _block == 0:
            _remainder = images[_blocks[_block + 1] - 500:_blocks[_block + 1], :, :].copy()
            images[0:_blocks[_block + 1], :, :] = cupy.asnumpy(fast_filter_images(cupy.asarray(
                images[0:_blocks[_block + 1], :, :]), Footprint))
        elif _block == _num_blocks - 1:

            images[_blocks[_block]:_total_frames, :, :] = \
                cupy.asnumpy(fast_filter_images(
                    cupy.asarray(np.append(_remainder, images[_blocks[_block]:_total_frames, :, :],
                                           axis=0)), Footprint))[block_buffer:, :, :]
        else:
            _remainder_new = images[_blocks[_block + 1] - 500:_blocks[_block + 1], :, :].copy()
            images[_blocks[_block]:_blocks[_block + 1], :, :] = \
                cupy.asnumpy(fast_filter_images(
                    cupy.asarray(np.append(_remainder, images[_blocks[_block]:_blocks[_block + 1], :, :],
                                           axis=0)), Footprint))[block_buffer:block_size+block_buffer, :, :]
            _remainder = _remainder_new.copy()

    return images
# REFACTOR + OPTIMIZE, MAKE OUTPUT CUPY.NDARRAY
# TODO UNIT TEST


@validate_longest_numpy_dimension(axis=0, pos=0)
def clean_image_stack(images: np.ndarray, artifact_length: int = 1000, stack_sizes: int = 7000) \
        -> np.ndarray:
    """
    Function to remove initial imaging frames such that any shutter artifact is removing and the resulting tensor
    is evenly divisible by the desired stack size

    :param images: images array (frames, y pixels, x pixels)
    :type images: numpy.ndarray
    :param artifact_length: number of frames considered artifact
    :type artifact_length: int = 1000
    :param stack_sizes: number of frames per stack
    :type stack_sizes: int = 7000
    :return: images
    :rtype: numpy.ndarray
    """
    _frames = images.shape[0]
    _crop_idx = _frames % stack_sizes
    if _crop_idx >= artifact_length:
        return images[_crop_idx:, :, :]
    else:
        _frames -= artifact_length
        _crop_idx = _num_frames % stack_sizes
        return images[artifact_length + _crop_idx:, :, :]
# TODO UNIT TEST


@validate_longest_numpy_dimension(axis=0, pos=0)
@validate_numpy_dimension_odd(odd_dimensions=(0, 1, 2), pos=1)
def fast_filter_images(images: np.ndarray, mask: np.ndarray = np.ones((3, 3, 3))) -> Any:
    """
    GPU-parallelized multidimensional median filter

    mask is of the form np.ones((frames, y pixels, x pixels)) with the origin in the center

    Requires CuPy

    :param images: image stack to be filtered (frames, y pixels, x pixels)
    :type images: numpy.ndarray
    :param mask: Mask of the median filter
    :type mask: numpy.ndarray = np.ones((3, 3, 3))
    :return: filtered_image (frames, y pixels, x pixels)
    :rtype: Any
    """
    return cupyx.scipy.ndimage.median_filter(cupy.asarray(images), footprint=mask)
# TODO UNIT TEST


@validate_longest_numpy_dimension(axis=0, pos=0)
@validate_numpy_dimension_odd(odd_dimensions=(0, 1, 2), pos=1)
def filter_images(images: np.ndarray, mask: np.ndarray = np.ones((3, 3, 3))) -> np.ndarray:
    """
    Denoise a tiff stack using a multidimensional median filter

    This function simply calls

    mask is of the form np.ones((frames, y pixels, x pixels) with the origin in the center

    :param images: images stack to be filtered (frames, y pixels, x pixels)
    :type images: numpy.ndarray
    :param mask: mask of the median filter
    :type mask: numpy.ndarray = np.ones((3, 3, 3))
    :return: filtered images (frames, y pixels, x pixels)
    :rtype: numpy.ndarray
    """
    return scipy.ndimage.median_filter(images, mask=mask)
# TODO UNIT TEST


@validate_longest_numpy_dimension(axis=0, pos=0)
def grouped_z_project(images: np.ndarray, bin_size: Union[Tuple[int, int, int], int],
                      function: Callable = np.mean) -> np.ndarray:
    """
    Utilize grouped z-project to downsample data

    Downsample example function -> np.mean

    :param images: A numpy array containing a tiff stack (frames, y pixels, x pixels)
    :type images: numpy.ndarray
    :param bin_size:  size of each bin passed to downsampling function
    :type bin_size: tuple[int, int, int] or tuple[int]
    :param function: group-z projecting function
    :type function: Callable = np.mean
    :return: downsampled image (frames, y pixels, x pixels)
    :rtype: numpy.ndarray
    """
    return skimage.measure.block_reduce(images, block_size=bin_size,
                                                     func=function).astype(images.dtype)
    # cast back down from float64
# TODO UNIT TEST



from __future__ import annotations
import numpy as np
from numba import jit
from scipy.signal import firwin, filtfilt
from typing import Optional

def calculate_dfof(traces: np.ndarray, frame_rate: float = 30, in_place: bool = False,
                   offset: float = 0.0, raw: Optional[np.ndarray] = None) \
        -> np.ndarray:
    """
    Calculates Δf/f0 (fold fluorescence over baseline). Baseline is defined as the 5th percentile of the signal
    after a 1Hz low-pass filter using a Hamming window.

    :param traces: matrix of traces in the form of neurons x frames
    :type traces: numpy.ndarray
    :param frame_rate: frame rate of dataset
    :type frame_rate: float = 30
    :param in_place: boolean indicating whether to perform calculation in-place
    :type in_place: bool = False
    :param offset: offset added to baseline; useful if traces are non-negative
    :type offset: float
    :param raw: raw dataset used to calculate baseline; useful if traces have been factorized
    :type raw: numpy.ndarray or None
    :return: Δf/f0 matrix of n neurons x m samples
    :rtype: numpy.ndarray
    """
    if in_place:
        dfof = traces
    else:
        dfof = traces.copy()

    taps = 30 # More taps mean higher frequency resolution, which in turn means narrower filters and/or steeper
    # roll‐offs.
    filter_frequency = 1 # (Hz)
    baseline_percentile = 5
    neurons, samples = dfof.shape

    # if for some reason sample is less than 90 frames we'll reduce the number of taps
    # we'll also make sure it's odd so we always have type I linear phase
    taps = min(taps, int(max(taps, samples/3)))
    if taps % 2 == 0:
        taps -= 1

    # determine padding length
    padding = 3 * taps

    # hamming window
    filter_window = firwin(taps, cutoff=filter_frequency, fs=frame_rate)

    for neuron in range(neurons):
        # filter
        filtered_trace = filtfilt(filter_window, [1.0], dfof[neuron, :], axis=0, padlen=padding)
        # calculate baseline
        baseline = np.percentile(filtered_trace, baseline_percentile, axis=0, keepdims=True)
        # calculate dfof
        dfof[neuron, :] = (dfof[neuron, :] - baseline) / baseline
    # TODO probably can do without loop, check scipy

    return dfof
# TODO DOCUMENT, UNIT TEST, IMPLEMENT RAW


@jit
def calculate_standardized_noise(fold_fluorescence_over_baseline: np.ndarray, frame_rate: float = 30) -> np.ndarray:
    """
    Calculates a frame-rate independent standardized noise as defined as:
        | :math:`v = \\frac{\sigma \\frac{\Delta F}F}\sqrt{f}`

    It is robust against outliers and approximates the standard deviation of Δf/f0 baseline fluctuations.
    For comparison, the more exquisite of the Allen Brain Institute's public datasets are approximately 1*%Hz^(-1/2)

    :param fold_fluorescence_over_baseline: fold fluorescence over baseline (i.e., Δf/f0)
    :type fold_fluorescence_over_baseline: numpy.ndarray
    :param frame_rate: frame rate of dataset
    :type frame_rate: float = 30
    :return: standardized noise (units are  1*%Hz^(-1/2) )
    :rtype: numpy.ndarray
    """
    return np.median(np.abs(np.diff(fold_fluorescence_over_baseline))) / np.sqrt(frame_rate)
# TODO UNIT TEST


@jit
def detrend_polynomial(traces: np.ndarray, in_place: bool = False) -> np.ndarray:
    """
    Detrend traces using a fourth-order polynomial

    :param traces: matrix of traces in the form of neurons x frames
    :type traces: numpy.ndarray
    :param in_place: boolean indicating whether to perform calculation in-place
    :type in_place: bool = False
    :return: detrended traces
    :rtype: numpy.ndarray
    """
    [_neurons, _samples] = traces.shape
    _samples_vector = np.arange(_samples)

    if in_place:
        detrended_matrix = traces
    else:
        detrended_matrix = traces.copy()

    for _neuron in range(_neurons):
        _fit = np.polyval(np.polyfit(_samples_vector, detrended_matrix[_neuron, :], deg=4), _samples_vector)
        detrended_matrix[_neuron] -= _fit
    return detrended_matrix
# TODO UNIT TEST


@jit
def _perona_malik_diffusion(trace: np.ndarray, iters: int = 5, kappa: int = 100, gamma: float = 0.15) -> np.ndarray:
    """
    Edge-preserving smoothing using perona malik diffusion

    :param traces: a matrix of neurons x frames
    :type traces: numpy.ndarray
    :param iters: number of iterations
    :type iters: int = 5
    :param kappa: diffusivity conductance
    :type kappa: int = 100
    :param gamma: step size (must be less than 1)
    :type gamma: float = 0.15
    :return: smoothed traces
    :rtype: numpy.ndarray
    """
    return
# TODO PASTE ME, DOCUMENT, UNIT TEST


def smooth_perona_malik(traces: np.ndarray, iters: int = 5, kappa: int = 100, gamma: float = 0.15) -> np.ndarray:
    """
    Edge-preserving smoothing using perona malik diffusion

    :param traces: a matrix of neurons x frames
    :type traces: numpy.ndarray
    :param iters: number of iterations
    :type iters: int = 5
    :param kappa: diffusivity conductance
    :type kappa: int = 100
    :param gamma: step size (must be less than 1)
    :type gamma: float = 0.15
    :return: smoothed traces
    :rtype: numpy.ndarray
    """
    return
# TODO PASTE ME, DOCUMENT, UNIT TEST



import os
import pytest
from shutil import rmtree
from CalSciPy.io_tools import load_all_tiffs, load_single_tiff, save_raw_binary, load_raw_binary, \
    save_single_tiff, save_tiff_stack, save_video, load_binary_meta
from CalSciPy.bruker import determine_bruker_folder_contents, repackage_bruker_tiffs, load_bruker_tiffs
import numpy as np
import pathlib
# noinspection PyProtectedMember
from PPVD.style import TerminalStyle

# noinspection DuplicatedCode
FIXTURE_DIR = "".join([os.getcwd(), "\\testing_data"])

DATASET = pytest.mark.datafiles(
    "".join([FIXTURE_DIR, "\\sample_datasets"]),
    keep_top_dir=False,
    on_duplicate="ignore",
)


def read_descriptions(file):
    return np.genfromtxt(str(file), delimiter=",", dtype="int")


@DATASET
def test_determine_bruker_folder_contents_passes(datafiles):
    for _dir in datafiles.listdir():

        _input_folder = next(pathlib.Path(_dir).glob("bruker_folder"))
        _descriptions = next(pathlib.Path(_dir).glob("description.txt"))
        _descriptions = read_descriptions(_descriptions)

        _contents = determine_bruker_folder_contents(_input_folder)

        for _test in enumerate(["Channel", "Plane", "Frame", "Height", "Width"]):
            assert _contents[_test[0]] == _descriptions[_test[0]], f"{TerminalStyle.GREEN}Description Mismatch: " \
                                                                   f"{TerminalStyle.YELLOW} failed on dataset " \
                                                                   f"{TerminalStyle.BLUE}{pathlib.Path(_dir).name}: " \
                                                                   f"{TerminalStyle.ORANGE}{_test[1]}" \
                                                                   f"{TerminalStyle.YELLOW} detection " \
                                                                   f"{TerminalStyle.RESET}"
        return

    rmtree(datafiles)


def test_determine_bruker_folder_contents_fails():
    with pytest.raises(ValueError):
        determine_bruker_folder_contents("")  # FAIL VALIDATE PATH NO ROOT/DRIVE
    with pytest.raises(ValueError):
        determine_bruker_folder_contents("C:\\&^6*")  # FAIL VALIDATE PATH PERMITTED CHARS
    with pytest.raises(FileNotFoundError):
        determine_bruker_folder_contents("C:\\673469346349673496734967349673205-3258-32856-3486")
        # FAIL VALIDATE PATH NOT FOUND
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        determine_bruker_folder_contents(125.6)
        # FAIL WRONG TYPE


@DATASET
def test_repackage_bruker_tiffs(datafiles, tmp_path):
    for _dir in datafiles.listdir():
        # INGEST
        _input_folder = next(pathlib.Path(_dir).glob("bruker_folder"))
        _descriptions = next(pathlib.Path(_dir).glob("description.txt"))
        _output_folder = "".join([str(tmp_path), "\\", str(pathlib.Path(_dir).stem), "_output"])
        # MAKE OUTPUT FOLDER
        os.mkdir(_output_folder)
        # RUN FUNCTION
        repackage_bruker_tiffs(_input_folder, _output_folder, (0, 0))
        # TEST
        _descriptions = read_descriptions(_descriptions)
        _contents = load_all_tiffs(_output_folder)
        assert _contents.shape[0] == np.cumprod(_descriptions[2])[-1], f"{TerminalStyle.GREEN}Image Mismatch: " \
                                                                       f"{TerminalStyle.YELLOW}failed on dataset" \
                                                                       f"{TerminalStyle.BLUE} " \
                                                                       f"{pathlib.Path(_input_folder).name} " \
                                                                       f"{TerminalStyle.RESET}"


def test_repackage_bruker_tiffs_fails():
    with pytest.raises(ValueError):
        repackage_bruker_tiffs("C:\\&^6", "C:\\&^6")  # FAIL PATH CHARACTERS
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        repackage_bruker_tiffs("C:\\", 125.6)  # FAIL PATH TYPE
    with pytest.raises(FileNotFoundError):
        repackage_bruker_tiffs("C:\\1381945328953298532895", "C:\\")  # FAIL PATH EXISTS


@DATASET
def test_load_bruker_tiffs(datafiles):
    for _dir in datafiles.listdir():
        # INGEST
        _input_folder = next(pathlib.Path(_dir).glob("bruker_folder"))
        _input_image = next(pathlib.Path(_dir).glob("Video_01_of_1.tif"))
        _descriptions = next(pathlib.Path(_dir).glob("description.txt"))
        # LOAD COMPARISON
        _descriptions = read_descriptions(_descriptions)
        _image1 = load_single_tiff(_input_image)
        # TEST
        _image2 = load_bruker_tiffs(_input_folder, 1, 0)[0]
        np.testing.assert_array_equal(_image1, _image2, err_msg=f"{TerminalStyle.GREEN}Image Mismatch: "
                                                                f"{TerminalStyle.YELLOW}failed on dataset"
                                                                f"{TerminalStyle.BLUE} "
                                                                f"{pathlib.Path(_input_folder).name} "
                                                                f"{TerminalStyle.RESET}")


def test_load_bruker_tiff_fails():
    with pytest.raises(ValueError):
        load_bruker_tiffs("")  # FAIL VALIDATE PATH NO ROOT/DRIVE
    with pytest.raises(ValueError):
        load_bruker_tiffs("C:\\&^6*")  # FAIL VALIDATE PATH PERMITTED CHARS
    with pytest.raises(FileNotFoundError):
        load_bruker_tiffs("C:\\673469346349673496734967349673205-3258-32856-3486")
        # FAIL VALIDATE PATH NOT FOUND
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        load_bruker_tiffs(125.6)
        # FAIL WRONG TYPE
